use std::ops::{Deref, DerefMut};
use std::time::Duration;

use bevy::math::vec2;
use bevy::prelude::*;
use iyes_loopless::prelude::ConditionSet;
use log::LevelFilter::Off;

use crate::system::flow::AppState;

#[derive(Clone, Debug, Component)]
pub struct FrameAnimation {
	frames: Vec<usize>,
	current_index: usize,
	frame_time: Duration,
	current_time: Duration,
}

impl FrameAnimation {
	pub fn new(frames: Vec<usize>, frame_time: Duration) -> Self {
		Self {
			current_index: 0,
			frames,
			frame_time,
			current_time: Duration::ZERO,
		}
	}
}

pub fn tick_frame_animations(
	time: Res<Time>,
	mut query: Query<(&mut FrameAnimation, &mut TextureAtlasSprite)>,
) {
	let dt = time.delta();
	for (mut animation, mut sprite) in &mut query {
		animation.current_time += dt;
		while animation.current_time > animation.frame_time {
			animation.current_index += 1;
			if animation.current_index == animation.frames.len() {
				animation.current_index = 0;
			}
			let new_time = animation.current_time - animation.frame_time;
			animation.current_time = new_time;
		}
		sprite.index = animation.frames[animation.current_index];
	}
}

#[derive(Clone, Debug, Component)]
pub struct Offset(Vec2);
impl From<Vec2> for Offset {
	fn from(other: Vec2) -> Self {
		Self(other)
	}
}
impl Deref for Offset {
	type Target = Vec2;
	fn deref(&self) -> &Self::Target {
		&self.0
	}
}
impl DerefMut for Offset {
	fn deref_mut(&mut self) -> &mut Self::Target {
		&mut self.0
	}
}

#[derive(Clone, Debug, Component)]
pub struct PositionTween {
	from: Vec2,
	to: Vec2,
	duration: Duration,
	elapsed: Duration,
}

impl PositionTween {
	pub fn new(from: Vec2, to: Vec2, duration: Duration) -> Self {
		Self {
			from,
			to,
			duration,
			elapsed: Duration::ZERO,
		}
	}

	pub fn get_value(&self) -> Vec2 {
		if self.elapsed == Duration::ZERO {
			self.to - self.from
		} else {
			(self.to - self.from) * (self.elapsed.as_secs_f32() / self.duration.as_secs_f32())
		}
	}
}

#[derive(Bundle)]
pub struct PositionBundle {
	position: PositionTween,
	offset: Offset,
}

impl From<PositionTween> for PositionBundle {
	fn from(other: PositionTween) -> Self {
		PositionBundle {
			offset: other.get_value().into(),
			position: other,
		}
	}
}

pub fn sync_tween_offset(time: Res<Time>, mut query: Query<(&mut PositionTween, &mut Offset)>) {
	let delta = time.delta();

	for (mut tween, mut offset) in &mut query {
		tween.elapsed = tween.duration.min(tween.elapsed + delta);
		let percent = tween.elapsed.as_secs_f32() / tween.duration.as_secs_f32();
		**offset = (tween.to - tween.from) * percent;
	}
}

pub fn tween_cleanup(mut commands: Commands, query: Query<(Entity, &PositionTween)>) {
	for (entity, position) in &query {
		if position.elapsed >= position.duration {
			commands
				.entity(entity)
				.remove::<PositionTween>()
				.remove::<Offset>();
		}
	}
}

pub struct AnimationsPlugin;
impl Plugin for AnimationsPlugin {
	fn build(&self, app: &mut App) {
		// app.add_system_set_to_stage(
		// 	CoreStage::PreUpdate,
		// 	ConditionSet::new()
		// 		.run_in_state(AppState::InGame)
		// 		.with_system(add_offset_to_tweened)
		// 		.into(),
		// )
		app.add_system_set_to_stage(
			CoreStage::Update,
			ConditionSet::new()
				.run_in_state(AppState::InGame)
				.with_system(sync_tween_offset)
				.with_system(tick_frame_animations)
				.into(),
		)
		.add_system_set_to_stage(
			CoreStage::Last,
			ConditionSet::new()
				.run_in_state(AppState::InGame)
				.with_system(tween_cleanup)
				.into(),
		);
	}
}
