use std::fmt::{Display, Formatter};

use bevy::prelude::*;
use iyes_loopless::prelude::ConditionSet;

use crate::entities::spawner::EntitySpawner;
use crate::entities::timing::GlobalTimer;
use crate::system::flow::AppState;

#[derive(Clone, Copy, Debug, Component)]
pub struct CurseText;

pub fn generate_curse(timer: Res<GlobalTimer>, mut events: EventWriter<CurseEvent>) {
	if timer.is_triggered() {
		let curse = match fastrand::usize(0..60) {
			0..=9 => CurseEvent::CurseOfAmnesia,
			10..=19 => CurseEvent::CurseOfPain,
			20..=29 => CurseEvent::CurseOfMonsters,
			30..=39 => CurseEvent::CurseOfWeakness,
			40..=49 => CurseEvent::CurseOfLoss,
			50..=59 => CurseEvent::CurseOfMidas,
			_ => CurseEvent::CurseOfPain,
		};

		events.send(curse);
	}
}

pub enum CurseEvent {
	CurseOfPain,
	CurseOfWeakness,
	CurseOfMonsters,
	CurseOfAmnesia,
	CurseOfMidas,
	CurseOfLoss,
}

impl Display for CurseEvent {
	fn fmt(&self, f: &mut Formatter<'_>) -> std::fmt::Result {
		write!(
			f,
			"{}",
			match self {
				CurseEvent::CurseOfPain => "Curse of Pain",
				CurseEvent::CurseOfWeakness => "Curse of Weakness",
				CurseEvent::CurseOfMonsters => "Curse of Monsters",
				CurseEvent::CurseOfAmnesia => "Curse of Amnesia",
				CurseEvent::CurseOfMidas => "Curse of Midas",
				CurseEvent::CurseOfLoss => "Curse of Loss",
			}
		)
	}
}

pub fn handle_curse(
	mut spawner: EntitySpawner,
	mut events: ResMut<Events<CurseEvent>>,
	query: Query<Entity, With<CurseText>>,
) {
	for event in events.drain() {
		for entity in &query {
			spawner.commands.entity(entity).despawn_recursive();
		}
		spawner.spawn_curse_text(event);
	}
}

pub struct CursesPlugin;
impl Plugin for CursesPlugin {
	fn build(&self, app: &mut App) {
		app.add_event::<CurseEvent>()
			.add_system_set_to_stage(
				CoreStage::PreUpdate,
				ConditionSet::new()
					.run_in_state(AppState::InGame)
					.with_system(generate_curse)
					.into(),
			)
			.add_system_set_to_stage(
				CoreStage::PostUpdate,
				ConditionSet::new()
					.run_in_state(AppState::InGame)
					.with_system(handle_curse)
					.into(),
			);
	}
}
