use bevy::math::uvec2;
use fastrand::Rng;

use crate::world::generators::utils::{draw_box, outline_floors, GenerationDirection};
use crate::world::generators::{MapGenerator, TMP_FLOOR_GROUP};
use crate::world::level_map::{Indexer, LevelMap, MapLayer, MapLayerType, MapTile};

pub struct Blobular;
impl MapGenerator for Blobular {
	fn generate(indexer: &Indexer, rng: &Rng) -> LevelMap {
		let width = indexer.width();
		let height = indexer.height();

		let mut floor_layer = vec![0; width * height];
		let mut wall_layer = vec![0; width * height];
		let mut decoration_layer = vec![0; width * height];

		let target_monsters = (indexer.width() * indexer.height()) / 25;
		let monster_attempts = target_monsters * 10;
		let mut entities = Vec::with_capacity(target_monsters);

		let initial_x = width / 2;
		let initial_y = width / 2;
		let initial_width = rng.usize(3..=5);
		let initial_height = rng.usize(3..=5);

		draw_box(
			initial_x - (6 / 2),
			initial_y - (4 / 2),
			6,
			4,
			TMP_FLOOR_GROUP,
			&mut floor_layer,
			&indexer,
		);

		draw_box(0, 0, 6, 10, TMP_FLOOR_GROUP, &mut floor_layer, &indexer);

		// draw_box(
		// 	initial_x - initial_width / 2,
		// 	initial_y - initial_height / 2,
		// 	initial_width,
		// 	initial_height,
		// 	TMP_FLOOR_GROUP,
		// 	&mut floor_layer,
		// 	&indexer,
		// );
		//
		for _ in 0..1000 {
			let initial = find_next_door(1000, &floor_layer, &rng, &indexer);
			if let Some((next_x, next_y, direction)) = initial {
				let size_selection = rng.usize(1..=10);
				let (room_width, room_height) = if size_selection < 8 {
					(rng.usize(3..=6), rng.usize(3..=6))
				} else {
					(rng.usize(7..=12), rng.usize(7..=12))
				};

				// Align room drawing to top left corner of new box
				match direction {
					GenerationDirection::Top => {
						let start_x = next_x - (room_width / 4);
						let start_y = next_y + room_height - 1;
						draw_box(
							start_x,
							start_y,
							room_width,
							room_height,
							TMP_FLOOR_GROUP,
							&mut floor_layer,
							&indexer,
						);
					}
					GenerationDirection::Bottom => {
						let start_x = next_x - (room_width / 4);
						let start_y = next_y;
						draw_box(
							start_x,
							start_y,
							room_width,
							room_height,
							TMP_FLOOR_GROUP,
							&mut floor_layer,
							&indexer,
						);
					}
					GenerationDirection::Left => {
						let start_x = next_x - (room_width - 1);
						let start_y = next_y;
						draw_box(
							start_x,
							start_y,
							room_width,
							room_height,
							TMP_FLOOR_GROUP,
							&mut floor_layer,
							&indexer,
						);
					}
					GenerationDirection::Right => {
						let start_x = next_x;
						let start_y = next_y;
						draw_box(
							start_x,
							start_y,
							room_width,
							room_height,
							TMP_FLOOR_GROUP,
							&mut floor_layer,
							&indexer,
						);
					}
					GenerationDirection::None => {}
				}
			}
		}

		outline_floors(&floor_layer, &mut wall_layer, indexer);

		LevelMap {
			spawn: uvec2(initial_x as u32, initial_y as u32),
			width,
			height,
			layers: vec![
				MapLayer::from_sized_list(
					width,
					height,
					floor_layer
						.into_iter()
						.map(|tile| match tile {
							0 => None,
							rest => Some(MapTile::new_floor(rest)),
						})
						.collect(),
					MapLayerType::Floor,
				),
				MapLayer::from_sized_list(
					width,
					height,
					wall_layer
						.into_iter()
						.map(|tile| match tile {
							0 => None,
							rest => Some(MapTile::new_wall(rest)),
						})
						.collect(),
					MapLayerType::Wall,
				),
				MapLayer::from_sized_list(
					width,
					height,
					decoration_layer
						.into_iter()
						.map(|tile| match tile {
							0 => None,
							rest => Some(MapTile::new_obstacle(rest)),
						})
						.collect(),
					MapLayerType::None,
				),
			],
			entities,
		}
	}
}

fn find_next_door(
	max_attempts: usize,
	floor: &Vec<usize>,
	rng: &Rng,
	idx: &Indexer,
) -> Option<(usize, usize, GenerationDirection)> {
	let mut remaining = max_attempts;
	let mut found = None;
	let mut last_direction = GenerationDirection::None;

	while remaining > 0 && found.is_none() {
		remaining = remaining.saturating_sub(1);

		let target_index = rng.usize(0..floor.len());
		let (x, y) = idx.reverse(target_index);

		// Edges of the map are unsuitable for new rooms
		if x == 0 || x == idx.width() - 1 || y == 0 || y == idx.height() - 1 {
			continue;
		}

		let mut adjacents = 0;

		// Check Left
		if floor[idx.index(x - 1, y)] != 0 {
			adjacents += 1;
			last_direction = GenerationDirection::Right;
		}
		// Check Right
		if floor[idx.index(x + 1, y)] != 0 {
			adjacents += 1;
			last_direction = GenerationDirection::Left;
		}
		// Check Top
		if floor[idx.index(x, y + 1)] != 0 {
			adjacents += 1;
			last_direction = GenerationDirection::Bottom;
		}
		// Check Bottom
		if floor[idx.index(x, y - 1)] != 0 {
			adjacents += 1;
			last_direction = GenerationDirection::Top;
		}

		if adjacents == 1 {
			found = Some((x, y, last_direction));
			break;
		}
	}

	found
}
